# -*- coding: utf-8 -*-
'''
Gazelib, a toolkit for gaze analysis.
'''
from gazelib.package import version as __version__  # noqa
import gazelib.containers  # noqa
import gazelib.io  # noqa
import gazelib.legacy  # noqa
import gazelib.settings  # noqa
import gazelib.statistics  # noqa
import gazelib.validation  # noqa
